`timescale 1ns / 1ps
// -----------------------------------------------------------------------------
//
// Copyright 2012(c) Analog Devices, Inc.
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//  - Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  - Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  - Neither the name of Analog Devices, Inc. nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//  - The use of this software may or may not infringe the patent rights
//    of one or more patent holders.  This license does not release you
//    from the requirement that you obtain separate licenses from these
//    patent holders to use this software.
//  - Use of the software either in source or binary form, must be run
//    on or directly connected to an Analog Devices Inc. component.
//
// THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// INTELLECTUAL PROPERTY RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// -----------------------------------------------------------------------------
// FILE NAME : AD7656_1.v
// MODULE NAME : AD7656_1
// AUTHOR : atofan
// AUTHOR’S EMAIL : alexandru.tofan@analog.com
// -----------------------------------------------------------------------------
// SVN REVISION: $WCREV$
// -----------------------------------------------------------------------------
// KEYWORDS : AD7656_1
// -----------------------------------------------------------------------------
// PURPOSE : Driver for the AD7656_1 6-Channel, 250 KSPS, 16-Bit Parallel ADCs
// -----------------------------------------------------------------------------
// REUSE ISSUES
// Reset Strategy      : Active high reset signal
// Clock Domains       : The design considered an 100 MHz input clock. The
// FPGA_CLOCK_FREQ parameter must be adjusted if the input clock is changed.
// Critical Timing     : The delay between CS and RD low and data read should
//                       be at least 36 / 45 ns. If intermediary logic is used,
//                       the additional delays must be considered.
// Test Features       : N/A
// Asynchronous I/F    : N/A
// Instantiations      : N/A
// Synthesizable (y/n) : Y
// Target Device       : Xilinx KC705 Kintex 7 Board / Microblaze embedded processor
// Other               : The driver reads data from the ADC in word mode.The
//                      result is forwarded to the upper module. The upper module
//                      knows which channel is read by using bits [18:16] from data_o
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
module AD7656_1
//----------- Ports Declarations -----------------------------------------------
    (
        // Clock and Reset signals
        input               fpga_clk_i,
        input               reset_n_i,
        // IP Control and Data Interface
        input      [2:0]    channel_i,
        output reg [18:0]   data_o,
        output reg          data_rd_ready_o,
        // AD7656_1 Control and Data Interface
        input      [15:0]   adc_db_i,
        input               adc_busy_i,
        output reg          adc_cs_n_o,
        output reg          adc_rd_n_o,
        output reg          adc_reset_o,
        output reg          adc_convst_a_o,
        output reg          adc_convst_b_o,
        output reg          adc_convst_c_o
    );

//------------------------------------------------------------------------------
//----------- Wire Declarations ------------------------------------------------
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//----------- Registers Declarations -------------------------------------------
//------------------------------------------------------------------------------
reg [8:0]       adc_current_state;
reg [8:0]       adc_next_state;
reg [31:0]      cycle_cnt;
reg [31:0]      delay_cs_rd_cnt;
reg [31:0]      rd_low_cnt;
reg [2:0]       channel_read;
reg [2:0]       data_remaining_cnt;
reg [2:0]       channel_suffix;

//------------------------------------------------------------------------------
//----------- Local Parameters -------------------------------------------------
//------------------------------------------------------------------------------
// ADC States
localparam  ADC_IDLE_STATE              = 9'b000000001;
localparam  ADC_START_CONV_STATE        = 9'b000000010;
localparam  ADC_WAIT_BUSY_HIGH_STATE    = 9'b000000100;
localparam  ADC_WAIT_BUSY_LOW_STATE     = 9'b000001000;
localparam  ADC_CS_LOW_STATE            = 9'b000010000;
localparam  ADC_RD_LOW_STATE            = 9'b000100000;
localparam  ADC_READ_DATA_STATE         = 9'b001000000;
localparam  ADC_SEND_DATA_STATE         = 9'b010000000;
localparam  ADC_WAIT_END_STATE          = 9'b100000000;

// ADC Timing
localparam real FPGA_CLOCK_FREQ         = 100;
localparam real ADC_CYCLE_TIME          = 4;
localparam real ADC_CS_RD_DELAY         = 0.070;
localparam real ADC_RD_LOW              = 0.040;
localparam [31:0]   ADC_CYCLE_CNT       = FPGA_CLOCK_FREQ * ADC_CYCLE_TIME - 2;
localparam [31:0]   ADC_DELAY_CS_RD_CNT = FPGA_CLOCK_FREQ * ADC_CS_RD_DELAY - 1;
localparam [31:0]   ADC_RD_LOW_CNT      = FPGA_CLOCK_FREQ * ADC_RD_LOW - 1;

//------------------------------------------------------------------------------
//----------- Initial Block ----------------------------------------------------
//------------------------------------------------------------------------------
initial
begin
    cycle_cnt       = ADC_CYCLE_CNT;
end

// Update the ADC timing counters count
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        cycle_cnt   <= ADC_CYCLE_CNT;
    end
    else
    begin
        if(adc_current_state == ADC_IDLE_STATE)
        begin
            cycle_cnt   <= ADC_CYCLE_CNT;
        end
        else if(cycle_cnt != 32'd0)
        begin
            cycle_cnt <= cycle_cnt - 32'd1;
        end
    end
end 

// Update the number of desired reads
always @(posedge fpga_clk_i)
begin
    if(adc_current_state == ADC_CS_LOW_STATE)
    begin
        data_remaining_cnt <= ((channel_i == 3'b001)||(channel_i == 3'b010)||(channel_i == 3'b100)) ? 3'd1 :
                              ((channel_i == 3'b011)||(channel_i == 3'b101)||(channel_i == 3'b110)) ? 3'd3 : 
                              3'd5;
    end
    else if(adc_current_state == ADC_SEND_DATA_STATE)
    begin
        data_remaining_cnt <= data_remaining_cnt - 3'd1;
    end
end

// Register States Process
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        adc_current_state <= ADC_IDLE_STATE;
    end
    else
    begin
        adc_current_state <= adc_next_state;
    end
end

// State Transition Process
always @(*)
begin
    adc_next_state <= adc_current_state;
    case(adc_current_state)
        ADC_IDLE_STATE:
            begin
                adc_next_state <= ADC_START_CONV_STATE;
            end
        ADC_START_CONV_STATE:
            begin
                adc_next_state <= ADC_WAIT_BUSY_HIGH_STATE;
            end
        ADC_WAIT_BUSY_HIGH_STATE:
            begin
                if((adc_busy_i == 1'b1)||(cycle_cnt == 32'd0))
                begin
                    adc_next_state <= ADC_WAIT_BUSY_LOW_STATE;
                end
            end
        ADC_WAIT_BUSY_LOW_STATE:
            begin
                if((adc_busy_i == 1'b0)||(cycle_cnt == 32'd0))
                begin
                    adc_next_state <= ADC_CS_LOW_STATE;
                end
            end
        ADC_CS_LOW_STATE:
            begin
                if(delay_cs_rd_cnt == 3'h0)
                begin
                    adc_next_state <= ADC_RD_LOW_STATE;
                end
            end
        ADC_RD_LOW_STATE:
            begin
                if(rd_low_cnt == 32'd0)
                begin
                    adc_next_state <= ADC_READ_DATA_STATE;
                end
            end
        ADC_READ_DATA_STATE:
            begin
                adc_next_state <= ADC_SEND_DATA_STATE;
            end
        ADC_SEND_DATA_STATE:
            begin 
                if(data_remaining_cnt == 3'd0)
                begin
                    adc_next_state <= ADC_WAIT_END_STATE;
                end
                else
                begin
                    adc_next_state <= ADC_RD_LOW_STATE;
                end
            end
        ADC_WAIT_END_STATE:
            begin
                if(cycle_cnt == 32'b0)
                begin
                    adc_next_state <= ADC_IDLE_STATE;
                end
            end
        default:
            begin
                adc_next_state <= ADC_IDLE_STATE;
            end
    endcase
end

// State Output Process
always @(posedge fpga_clk_i)
begin
    if(reset_n_i == 1'b0)
    begin
        adc_cs_n_o      <= 1'b1;
        adc_rd_n_o      <= 1'b1;
        adc_convst_a_o  <= 1'b0;
        adc_convst_b_o  <= 1'b0;
        adc_convst_c_o  <= 1'b0;
        data_rd_ready_o <= 1'b0;
        adc_reset_o     <= 1'b1;
        channel_suffix  <= 3'd0;
    end
    else
    begin
        case(adc_current_state)
            ADC_IDLE_STATE:
                begin
                    delay_cs_rd_cnt <= ADC_DELAY_CS_RD_CNT;
                    rd_low_cnt      <= ADC_RD_LOW_CNT;
                    adc_cs_n_o      <= 1'b1;
                    adc_rd_n_o      <= 1'b1;
                    adc_convst_a_o  <= 1'b0;
                    adc_convst_b_o  <= 1'b0;
                    adc_convst_c_o  <= 1'b0;
                    data_rd_ready_o <= 1'b0;
                    adc_reset_o     <= 1'b0;
                    channel_suffix  <= (channel_i[0] == 1'b1) ? 3'd0 :
                                       (channel_i[1] == 1'b1) ? 3'd2 :
                                       (channel_i[2] == 1'b1) ? 3'd4 :
                                       3'd0;
                end
            ADC_START_CONV_STATE:
                begin 
                    adc_cs_n_o      <= 1'b1;
                    adc_rd_n_o      <= 1'b1;
                    adc_convst_a_o  <= channel_i[0];
                    adc_convst_b_o  <= channel_i[1];
                    adc_convst_c_o  <= channel_i[2];
                    data_rd_ready_o <= 1'b0;
                    adc_reset_o     <= 1'b0;
                end
            ADC_WAIT_BUSY_HIGH_STATE:
                begin
                    adc_cs_n_o      <= 1'b1;
                    adc_rd_n_o      <= 1'b1;
                    adc_convst_a_o  <= channel_i[0];
                    adc_convst_b_o  <= channel_i[1];
                    adc_convst_c_o  <= channel_i[2];
                    data_rd_ready_o <= 1'b0;
                    adc_reset_o     <= 1'b0;
                end
            ADC_WAIT_BUSY_LOW_STATE:
                begin
                    adc_cs_n_o      <= 1'b1;
                    adc_rd_n_o      <= 1'b1;
                    adc_convst_a_o  <= channel_i[0];
                    adc_convst_b_o  <= channel_i[1];
                    adc_convst_c_o  <= channel_i[2];    
                    data_rd_ready_o <= 1'b0;  
                    adc_reset_o     <= 1'b0;                    
                end
            ADC_CS_LOW_STATE:
                begin
                    delay_cs_rd_cnt <= delay_cs_rd_cnt - 32'd1;
                    adc_cs_n_o      <= 1'b0;
                    adc_rd_n_o      <= 1'b1;
                    adc_convst_a_o  <= 1'b0;
                    adc_convst_b_o  <= 1'b0;
                    adc_convst_c_o  <= 1'b0;
                    data_rd_ready_o <= 1'b0;    
                    adc_reset_o     <= 1'b0;
                end
            ADC_RD_LOW_STATE:
                begin
                    rd_low_cnt      <= rd_low_cnt - 32'd1;
                    adc_cs_n_o      <= 1'b0;
                    adc_rd_n_o      <= 1'b0;
                    adc_convst_a_o  <= 1'b0;
                    adc_convst_b_o  <= 1'b0;
                    adc_convst_c_o  <= 1'b0;  
                    data_rd_ready_o <= 1'b0;
                    adc_reset_o     <= 1'b0;
                end
            ADC_READ_DATA_STATE:
                begin
                    adc_cs_n_o      <= 1'b0;
                    adc_rd_n_o      <= 1'b0;
                    adc_convst_a_o  <= 1'b0;
                    adc_convst_b_o  <= 1'b0;
                    adc_convst_c_o  <= 1'b0;
                    data_rd_ready_o <= 1'b0;
                    adc_reset_o     <= 1'b0;    
                    data_o          <= {channel_suffix, adc_db_i};
                end
            ADC_SEND_DATA_STATE:
                begin
                    rd_low_cnt      <= ADC_RD_LOW_CNT;
                    adc_cs_n_o      <= 1'b0;
                    adc_rd_n_o      <= 1'b1;
                    adc_convst_a_o  <= 1'b0;
                    adc_convst_b_o  <= 1'b0;
                    adc_convst_c_o  <= 1'b0;
                    data_rd_ready_o <= 1'b1; 
                    adc_reset_o     <= 1'b0;
                    channel_suffix  <= ((channel_i[0] == 1'b1)&&(channel_suffix < 3'd1)) ? (channel_suffix + 3'd1) :
                                       ((channel_i[1] == 1'b0)&&(channel_suffix < 3'd3)) ? (channel_suffix + 3'd3) : 
                                       ((channel_i[1] == 1'b1)&&(channel_suffix < 3'd3)) ? (channel_suffix + 3'd1) :
                                       ((channel_i[2] == 1'b1)&&(channel_suffix < 3'd5)) ? (channel_suffix + 3'd1) : 3'd0;
                end
            ADC_WAIT_END_STATE:
                begin
                    adc_cs_n_o      <= 1'b1; 
                    adc_rd_n_o      <= 1'b1;
                    adc_convst_a_o  <= 1'b0;
                    adc_convst_b_o  <= 1'b0;
                    adc_convst_c_o  <= 1'b0; 
                    data_rd_ready_o <= 1'b0;
                    adc_reset_o     <= 1'b0;
                end 
            default:
                begin
                    adc_cs_n_o      <= 1'b1;
                    adc_rd_n_o      <= 1'b1;
                    adc_convst_a_o  <= 1'b0;
                    adc_convst_b_o  <= 1'b0;
                    adc_convst_c_o  <= 1'b0;
                    data_rd_ready_o <= 1'b0;
                    adc_reset_o     <= 1'b1;
                    channel_suffix  <= 3'd0;
                end
        endcase
    end
end

endmodule
